// 	KEDS Project                                                     OneADayFilter.c																											(Kansas Event Data System)
//__________________________________________________________________________________
// 																																	OneADayFilter.c
/*

This program filters event data using the rule that each dyad can have one only 
event per coding category per day ("daily unique dyad-code rule").  This algorithm 
eliminates all event coded from duplicate stories at the expense of a few false 
positives.

USER INPUT

The program asks for the name of the file to be filtered -- this should be followed
by <rtn>.  It also asks whether a discard file should be generated; response with
Y (yes) or N (no) followed by <rtn>.

SCREEN FEEDBACK

Program currently shows the Julian date (1 Jan 1960 = 0) that is being evaluated,
then a string of characters showing the status of each record on that date.  This
display slows the program significantly, and can be eliminated by commenting-out
the appropriate printf() and putchar() statements.

INPUT FILE

The input file is standard TABARI formatted event data of the form

YYMMDD \t SRC \t TAR \t EVT (other text)

where
	\t       tab
	YYMMDD   6-character date; 00 through 39 are windowed to 2000 to 2039
	SRC      code for source actor (3 to 6 characters)	
	TAR      code for target actor (3 to 6 characters)	
	EVT      code for event (2 to 4 characters)
	other    up to 230 characters of additional text; this is ignored in the 
					 evaluation but copied to the new file
	
The input format is used only in the main() function	

OUTPUT FILES

The output has the name of the input file followed by the suffix ".filt".  It is 
identical in format to the input file and simply contains copies of the retained
records.

An optional discard file -- input file name followed by the suffix ".disc" -- 
contains the discarded records; this can be used to confirm that the program is, 
in fact, generally discarding junk.  The first characters -- src, tar, evt --
indicate the reason for the discard.

Notes:

0. This program is based on a far more complicated "Bayesian filter", EventFilter.c,
   that tracked events over a longer period of time (months, years) and was used
   to eliminate low-probability codes such a Switzerland attacking the USA.  It
   is perhaps overkill for single-day filtering, since a much simpler perl
   program (read a day of data, sort, eliminate records that duplicate the first
   four fields) would accomplish the same thing. But it works...
   
1. The core data structure for the program are the three linked-lists stored in
   the structures SrcListStruct, TarListStruct, and EvtListStruct.  These form
   a hierarchy based on source code, target code, and event code respectively --
   in other words, each source points to a list of targets and each of those 
   (now a source-target combination) points to a list of events. 
   
   The source and target lists are dynamically sorted by frequency: the [untested] 
   logic of this is that event data are highly asymmetrically distributed, and 
   therefore the pay-off for finding frequent items early in the list is higher than 
   other possible sortings of the list (e.g. alphabetical).
      
   On the assumption that time is more important than memory, the new items can be
   added to the lists (in the function AddRecord) but they are not deleted even 
   when the frequency drops to zero.  This avoids the garbage collection and 
   continual reallocation of memory that would otherwise occur with low-frequency
   event combinations.  If you need to conserve memory, deletion of links could
   be done in the function DeleteRecord().
   
   [Note: the current algorithm does not properly swap the second record.  Ran out 
   of time before I could fix this; it does not affect the filtering.]
   
3. This version uses 3-character actor codes and 3 or 4-character event codes; for
   purposes of comparison, 3-character event codes are blank-terminated.  These  
   limits are hard-coded but it would be fairly straightforward to generalize them.
   The storage routine TransStore() and the string-truncation in GetRecord() will
   need to be modified, as well as the storage array sizes.  The input routine 
   assumes that actor codes have no more than 6 characters, and event codes have no 
   more than 4 characters.

4. An assortment of errors are trapped but without recovery: program will give the 
	 truncated content of the record that caused the error but then quit.  Assumption 
	 is that this is being run by someone who has access to the source code and can
	 make the appropriate corrections...   

5. Y2K Compliance: This software uses a YYMMDD date format and then "windows" 
	 two-digit years from 00 through 39 to 2000 to 2039.  The program uses Stata, Inc.
	 Julian accounting, which begins 1 Jan 1960, so dates are assumed to begin after 
	 that (with current event data, series usually start around 1980 or 1990).  If 
	 additional windowing is needed, modify the function JulianStata();
	 
SYSTEM REQUIREMENTS:
This program was originally developed on the Macintosh operating system (OS 9)
but contains no Macintosh-specific features and will presumably work on any
system that has a C compiler. Current version was compiled under Macintosh
OS-10.3

Compiler:		original: Metrowerks C
						revised:  gcc
Language:		ANSI C
For plausible indenting of this source code, set the tab size in your editor to "2"

REVISION HISTORY
  June-02:  Initial version, based on EventFilter.c
  June-05:  Recompiled in gcc

PROVENANCE
Programmer: Philip A. Schrodt
            Kansas Event Data System Project
            Center for International Political Analysis
            University of Kansas 
            Lawrence, Kansas, 66044 U.S.A.
            http://www.ku.edu/~keds
			email: schrodt@ku.edu
						
Copyright (c) 2002  Philip A. Schrodt.  All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted under the terms of the GNU General Public License:
http://www.opensource.org/licenses/gpl-license.html

Report bugs to: schrodt@ku.edu

*/

//__________________________________________________________________________________

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//__________________________________________________________________________________
// 	Macros

#define	FALSE			0
#define TRUE			1

#define IN_FILE		"LVNT.WEIS.79-89" // Debugging
#define MAX_FILE_NAME	32			// maximum length of file name 
#define FILT_SUFFIX	".filt"		// suffix for the filtered file
#define DISC_SUFFIX	".disc"  	// suffix for file with global frequency information

#define MAXSTORE  2000				// maximum size of record storage array	
#define MAX_INPUT_LINE   256	// maximum size of event record
#define MAX_ACTOR_LENGTH   7	// maximum length (plus 1) of actor codes
#define MAX_EVENT_LENGTH   5	// maximum length (plus 1) of event codes
#define MAX_DATE_LENGTH    7	// maximum length (plus 1) of dates


//__________________________________________________________
// Structures

struct evtListStruct { 	 // event list elements
	char evt[3];				//  event code
	int	 efreq;					// frequency
	struct evtListStruct * pnext;  // next list item
};

struct tarListStruct { 	 // target list elements
	char tactor[4];				// actor code
	int	 tfreq;						// frequency
	struct evtListStruct * pevt;  // start of event list
	struct tarListStruct * pnext;  // next list item
};

struct srcListStruct { 	 // source list elements
	char sactor[4];				// actor code
	int	 sfreq;						// frequency
	struct tarListStruct * ptar;  // start of target list
	struct srcListStruct * pnext;  // next list item
}  * srcListHead;


//__________________________________________________________
// Globals

FILE	*fin, *ffilt, *fdisc;				// input and output files
char  sfin_name[MAX_FILE_NAME], sfilt_name[MAX_FILE_NAME]; // names of same
int		doDiscard = TRUE;			// write discard file?
  
int		Month[] = {0,0,31,59,90,120,151,181,212,243,273,304,334};		// used for Julian computation: # days prior to Month[n] JAN=1

char  src[MAX_ACTOR_LENGTH] = "XXX";	// source code 
char  tar[MAX_ACTOR_LENGTH] = "XXX";	// target code 
char  evt[MAX_EVENT_LENGTH] = "----";	// event code 
char  date[MAX_DATE_LENGTH];					// character version of date of current record
char	sinput[MAX_INPUT_LINE];					// read input line
char  olddate[7] = "000000";					// character version of date of record from current day
char  c;

//___________________________________________________________________________
// Prototypes

void Initialize();  					// initialize globals
void GetFileNames();  				// get names of files from keyboard
int  GetRecord(void);					// read record from fin; returns false if feof(fin)
void NewTarget(struct tarListStruct ** ptar); // Creates new target and transfers information
void NewEvent(struct evtListStruct ** pevt); 	// Transfers information to new event link 
int  AddRecord(void);  				// add record information; returns TRUE is new dyad-type
void Report_Error(char s[]);  // exit on error
void ResetCount(void);  			// zero the dyad-code counts

//__________________________________________________________________________________
//																																			main program

int main(void)
{
Initialize();

GetFileNames();

while (GetRecord())  
{
	if (strcmp(date,olddate)) {	// new day
		printf("Filtering %s\n: ",date);
		ResetCount();  	// zero the dyad-code counts
		strcpy(olddate,date);
	}
	if (AddRecord())    fputs(sinput,ffilt);
	else if (doDiscard) fputs(sinput,fdisc);
}
			
putchar('\n');
fclose(fin);
fclose(ffilt);
if (doDiscard) fclose(fdisc);
printf("\aFinished!\n");
printf("Press Cmd-Q to quit");  // CodeWarrior specific
	
} // main

void Initialize()
// initialize globals
{
// initialize list head
	srcListHead =  malloc(sizeof(struct srcListStruct));
	if (NULL == srcListHead) Report_Error("Insufficient memory to initialize srcListHead");
	strcpy(srcListHead->sactor,"XXX");			// put something here so that we don't need to make an exception for the empty list
//	srcListHead->sfreq = 0;
	srcListHead->pnext = NULL;
	strcpy(tar,"XXX");
	NewTarget(&srcListHead->ptar);
	strcpy(evt,"00");
	NewEvent(&srcListHead->ptar->pevt);
	
} // Initialize

void GetFileNames()  
// get names of files from keyboard; also open temporary file and initialize adate
{
#if TRUE
	printf("\n=========================================================================\n\n");
	printf("OneADay_Filter: Program for filtering events from a stream of event  data\n");
	printf("                using a daily unique dyad-code rule (Version 0.5B2)\n");
	printf("                Kansas Event Data System (KEDS)\n");
	printf("\n(c) 2002, University of Kansas\n");
	printf("Redistribution and use in source and binary forms, with or without modification,\n");
	printf("are permitted under the terms of the GNU General Public License:\n");
	printf("http://www.opensource.org/licenses/gpl-license.html\n");
	printf("\nPress <return> to start\n");
	gets(src);
#endif

#if TRUE					// get names from keyboard
	printf("\aEnter name of file to be filtered:");
	gets(sfin_name);
	if (strlen(sfin_name) >= MAX_FILE_NAME - 6)  Report_Error("Input file name is too long to allow \".filt\" suffix to be added ");
	
	printf("Write discarded events to file? (Y/N):");
	gets(sinput);
	if (('n' == sinput[0]) || ('N' == sinput[0])) doDiscard = FALSE;
	else doDiscard = TRUE;

#else							// get names from macros (debugging)
	strcpy(sfin_name,IN_FILE);
#endif
	fin = fopen(sfin_name,"r");
	if (!fin) Report_Error("Unable to open input file");
	
// initialize curdate
	fgets(sinput,MAX_INPUT_LINE,fin);
	sinput[6] = '\0';  // just look at the date
	strcpy(olddate,sinput);
	rewind(fin);
		
	strcpy(sfilt_name,sfin_name);  // create .filt file name and open file
	strcat(sfilt_name,FILT_SUFFIX);
	ffilt = fopen(sfilt_name,"w");
	printf("Filtered events are in the file \"%s\"\n",sfilt_name);
	if (!ffilt) Report_Error("Unable to open filtered event file");
	if (doDiscard) {  // create .discard file name and open file
		strcpy(sfilt_name,sfin_name);
		strcat(sfilt_name,DISC_SUFFIX);
		fdisc = fopen(sfilt_name,"w");
		printf("Discarded events are in the file \"%s\"\n",sfilt_name);
		if (!fdisc) Report_Error("Unable to open discarded event file");
	}

} // GetFileNames

int GetRecord(void)
// read record from fin; returns false if feof(fin)
{
	fgets(sinput,MAX_INPUT_LINE,fin);
	if (feof(fin)) return FALSE;
	sscanf(sinput,"%s %s %s %s",&date, &src, &tar, &evt);
	src[3] = '\0';  // truncate the actor codes
	tar[3] = '\0';
	if (!evt[3]) { // add terminal ' ' to 3-character codes
		evt[3] = ' ';	
		evt[4] = '\0';
	}	
	return TRUE;
} // GetRecord


void NewTarget(struct tarListStruct ** ptar)
// Transfers information to new target link 
{
	*ptar =  malloc(sizeof(struct tarListStruct));
	if (NULL == *ptar) Report_Error("Insufficient memory for new target");
	strcpy((*ptar)->tactor,tar);
	(*ptar)->tfreq = 1;
	(*ptar)->pnext = NULL;
} // NewTarget

void NewEvent(struct evtListStruct ** pevt)
// Transfers information to new event link 
{
	*pevt =  malloc(sizeof(struct evtListStruct));
	if (NULL == *pevt) Report_Error("Insufficient memory for new event");
	strcpy((*pevt)->evt,evt);
	(*pevt)->efreq = 1;
	(*pevt)->pnext = NULL;
} // NewTarget

int AddRecord(void)
// Adds record to cstore; returns TRUE if dyad-code is new
{
char * pc;
char * pca;
struct srcListStruct * psLS,* ppsLS, * psrc; 
struct tarListStruct * ptLS,* pptLS, * ptar;
struct evtListStruct * peLS,* ppeLS;

int ka; // debug: used to pause in list display
	
	ppsLS = NULL;
	psLS = srcListHead;
	while (psLS && strcmp(src,psLS->sactor)) { // traverse list looking for src
		ppsLS = psLS;
		psLS = psLS->pnext;
	}

	if (!psLS) { // got to end of list without finding actor; make a new entry
		ppsLS->pnext = malloc(sizeof(struct srcListStruct));
		if (NULL == ppsLS->pnext) Report_Error("Insufficient memory for new source");
		psLS = ppsLS->pnext;
		strcpy(psLS->sactor,src);
		psLS->sfreq = 1;
		psLS->pnext = NULL;
		NewTarget(&psLS->ptar);
		NewEvent(&psLS->ptar->pevt);
		return TRUE;
	}
	else {  // found it; increment counter and set ptLS to head of list
			++psLS->sfreq;
			if (ppsLS && (psLS->sfreq > ppsLS->sfreq)) { // swap records if frequency is higher than predecessors
			psrc = srcListHead;
			while (psrc && (psrc->pnext != ppsLS)) psrc = psrc->pnext;  // find pointer to ppsLS
			if (psrc) {
				psrc->pnext = psLS; // swap
				ppsLS->pnext = psLS->pnext;
				psLS->pnext = ppsLS;
			}
		}
			
		ptLS = psLS->ptar;
		pptLS = NULL;
		if (strcmp(tar,ptLS->tactor)) { // find target 
			pptLS = ptLS;
			ptLS = pptLS->pnext;
			while (ptLS && strcmp(tar,ptLS->tactor)) { // traverse list looking for tar
				pptLS = ptLS;
				ptLS = ptLS->pnext;
			}
		}
		if (!ptLS) { // got to end of list without finding actor; make a new entry
			NewTarget(&pptLS->pnext);
			NewEvent(&pptLS->pnext->pevt);
			return TRUE;
		}
		else {  // found it; increment counter
				++ptLS->tfreq;
				if (pptLS && (ptLS->tfreq > pptLS->tfreq)) { // swap records if frequency is higher than predecessors
				ptar = psLS->ptar;
				while (ptar && (ptar->pnext != pptLS)) ptar = ptar->pnext;  // find pointer to ppsLS
				if (ptar) {
					ptar->pnext = ptLS; // swap
					pptLS->pnext = ptLS->pnext;
					ptLS->pnext = pptLS;
				}
			}
			peLS = ptLS->pevt;
			ppeLS = NULL;
			if (strcmp(evt,peLS->evt)) { // find event 
				ppeLS = peLS;
				peLS = peLS->pnext;
				while (peLS && strcmp(evt,peLS->evt)) { // traverse list looking for event
					ppeLS = peLS;
					peLS = peLS->pnext;
				}
			}
			if (!peLS) { // got to end of list without finding event; make a new entry
				NewEvent(&ppeLS->pnext);
				return TRUE;
			}
			else {  // found it; increment counter
				if (peLS->efreq++) return FALSE;
				else return TRUE;
			}
		}
	}

#if FALSE
// ### debug: display list
	psLS = srcListHead->pnext;
//	psLS = srcListHead;   // first entry is currently meaningless... 
	while (psLS) {
		printf("%s  %d\n",psLS->sactor,psLS->sfreq);
		ptLS = psLS->ptar;
		while (ptLS) {
			printf("    %s %d:",ptLS->tactor,ptLS->tfreq);
			peLS = ptLS->pevt;
			while (peLS) {
				printf(" %s(%d)",peLS->evt ,peLS->efreq);
				peLS = peLS->pnext;
			}	
			ptLS = ptLS->pnext;
			printf("\n");
		}		
		psLS = psLS->pnext;
	}
	ka = getchar();  // pause until return
#endif
	
} // AddRecord

void Report_Error(char s[])  
// exit on early end-of-file without recovery
{
	printf("\a%s\nLast line read was\n%s %s %s %s\n",s,date,src,tar,evt);
	printf("Press <return> to exit");
	gets(src);
	exit(0);
} // Error_Exit


void ResetCount(void)
// zero the dyad-code counts
{
struct srcListStruct * psLS; 
struct tarListStruct * ptLS;
struct evtListStruct * peLS;
// ### debug: display list
	psLS = srcListHead->pnext;
//	psLS = srcListHead;   // first entry is currently meaningless... 
	while (psLS) {
//		printf("%s  %d\n",psLS->sactor,psLS->sfreq);
		ptLS = psLS->ptar;
		while (ptLS) {
//			printf("    %s %d:",ptLS->tactor,ptLS->tfreq);
			peLS = ptLS->pevt;
			while (peLS) {
//				printf(" %s(%d)",peLS->evt ,);
				peLS->efreq = 0;
				peLS = peLS->pnext;
			}	
			ptLS = ptLS->pnext;
//			printf("\n");
		}		
		psLS = psLS->pnext;
	}
		
} // ResetCount
	
